package.loaded.Constants = nil; require("Constants")

CrabBehaviors = {}

-- move to the next waypoint
function CrabBehaviors.GoToWpt(self)
	if not self.MOMoveTarget then
		local Destination = self:GetLastAIWaypoint()
		if SceneMan:ShortestDistance(Destination, self.Pos, false).Magnitude < 20 then
			self:ClearAIWaypoints()
			if self.AIMode == Actor.AIMODE_GOTO then
				self.AIMode = Actor.AIMODE_SENTRY
			end
			return true
		end
	end
	
	local UpdatePathTimer = Timer()
	UpdatePathTimer:SetSimTimeLimitMS(5000)
	
	local StuckTimer = Timer()
	StuckTimer:SetSimTimeLimitMS(2000)
	
	local PrevWptPos = self.Pos
	local WptList, Waypoint, Dist, CurrDist
	
	while true do
		while self.Target and self.FirearmIsReady do	-- don't move around if we have something to shoot at
			coroutine.yield()	-- wait until next frame
		end
		
		if self.Vel.Magnitude > 2 then
			StuckTimer:Reset()
		end
		
		if self.MOMoveTarget then	-- make the last waypoint marker stick to the MO we are following
			if MovableMan:ValidMO(self.MOMoveTarget) then
				self:RemoveMovePathEnd()
				self:AddToMovePathEnd(self.MOMoveTarget.Pos)
			else
				self.MOMoveTarget = nil
			end
		end
		
		if UpdatePathTimer:IsPastSimTimeLimit() then
			UpdatePathTimer:Reset()
			self.deviceState = AHuman.STILL
			self.lateralMoveState = Actor.LAT_STILL
			Waypoint = nil
			WptList = nil
		elseif StuckTimer:IsPastSimTimeLimit() then	-- dislodge
			StuckTimer:Reset()
			if self.lateralMoveState == Actor.LAT_LEFT then
				self.lateralMoveState = Actor.LAT_RIGHT
			elseif self.lateralMoveState == Actor.LAT_LEFT then
				self.lateralMoveState = Actor.LAT_LEFT
			else
				self.lateralMoveState = math.random(Actor.LAT_LEFT, Actor.LAT_RIGHT)
			end
		elseif WptList then	-- we have a list of waypoints, folow it
			if #WptList < 1 and not Waypoint then	-- arrived
				break
			else
				if not Waypoint then	-- get the next waypoint in the list
					-- if we are following someone we must update the path more often
					if self.MOMoveTarget and MovableMan:ValidMO(self.MOMoveTarget) and 
						 SceneMan:ShortestDistance(self:GetLastAIWaypoint(), self.MOMoveTarget.Pos, false).Magnitude > 40 
					then	-- the target has moved
						WptList = nil
						Waypoint = nil
					else
						UpdatePathTimer:Reset()
						Waypoint = table.remove(WptList, 1)
						if #WptList > 0 then
							self:RemoveMovePathBeginning()
						elseif SceneMan:ShortestDistance(self.Pos, Waypoint.Pos, false).X < 10 then	-- the last waypoint
							self:ClearAIWaypoints()
							self.AIMode = Actor.AIMODE_SENTRY
							break
						end
					end
				end
				
				if Waypoint then
					if self.MOMoveTarget and MovableMan:ValidMO(self.MOMoveTarget) and
						SceneMan:ShortestDistance(self.Pos, self.MOMoveTarget.Pos, false).Magnitude < (self.Radius+self.MOMoveTarget.Radius)*3
					then
						self.lateralMoveState = Actor.LAT_STILL
					else
						CurrDist = SceneMan:ShortestDistance(self.Pos, Waypoint.Pos, false)
						if CurrDist.X < -3 then
							self.lateralMoveState = Actor.LAT_LEFT
						elseif CurrDist.X > 3 then
							self.lateralMoveState = Actor.LAT_RIGHT
						else
							Waypoint = nil
						end
					end
				end
			end
		else	-- no waypoint list, create one in several small steps to reduce lag
			local TmpList = {}
			table.insert(TmpList, {Pos=self.Pos})
			self:UpdateMovePath()
			coroutine.yield()	-- wait until next frame
			
			for WptPos in self.MovePath do	-- skip any waypoint too close to the previous one
				if SceneMan:ShortestDistance(TmpList[#TmpList].Pos, WptPos, false).Magnitude > 10 then
					table.insert(TmpList, {Pos=WptPos})
				end
			end
			
			if #TmpList < 3 then
				Dist = nil
				if TmpList[2] then
					Dist = SceneMan:ShortestDistance(TmpList[2].Pos, self.Pos, false)
				end
				
				-- already at the target
				if not Dist or Dist.Magnitude < 25 then
					self:ClearMovePath()
					break
				end
			end
			
			coroutine.yield()	-- wait until next frame
			
			WptList = TmpList
			
			-- create the move path seen on the screen
			self:ClearMovePath()
			for _, Wpt in pairs(TmpList) do
				self:AddToMovePathEnd(Wpt.Pos)
			end
		end
		
		if self.BlockingActor then
			if not MovableMan:IsActor(self.BlockingActor) or
				SceneMan:ShortestDistance(self.Pos, self.BlockingActor.Pos, false).Magnitude > (self.Diameter + self.BlockingActor.Diameter)*1.2
			then
				self.BlockingActor = nil
				self.teamBlockState = Actor.NOTBLOCKED
				
				if self.AIMode == Actor.AIMODE_BRAINHUNT and self.FollowingActor then
					self.FollowingActor = nil
					break	-- end this behavior
				end
			elseif self.teamBlockState == Actor.NOTBLOCKED and Waypoint then
				if (Waypoint.Pos.X > self.Pos.X and self.BlockingActor.Pos.X > self.Pos.X) or
					 (Waypoint.Pos.X < self.Pos.X and self.BlockingActor.Pos.X < self.Pos.X)
				then
					self.teamBlockState = Actor.BLOCKED
					if self.AIMode == Actor.AIMODE_BRAINHUNT and self.AIMode == self.BlockingActor.AIMode and
						(not self.BlockingActor.MOMoveTarget or self.BlockingActor.MOMoveTarget.ID ~= self.ID)
					then	-- don't follow an actor that is following us
						self.FollowingActor = self.BlockingActor
						self:ClearAIWaypoints()
						self:AddAIMOWaypoint(self.FollowingActor)
						self:CreateGoToBehavior()
					end
				else
					self.BlockingActor = nil
				end
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	return true
end

-- stop the user fom inadvertently modifying the storage table
local Proxy = {}
local Mt = {
	__index = CrabBehaviors,
	__newindex = function(Table, k, v)
		error("The CrabBehaviors table is read-only.", 2)
	end
}
setmetatable(Proxy, Mt)
CrabBehaviors = Proxy
