package.loaded.Constants = nil; require("Constants")
require("Actors/AI/PID")

HumanBehaviors = {}

-- finds an aming angle that compensate for projectile drop caused by gravity
function HumanBehaviors.GetHitAngle(AimPoint, TargetVel, StartPos, muzVel)
	local Dist = SceneMan:ShortestDistance(StartPos, AimPoint, false)
	local range = Dist.Magnitude
	
	-- compensate for gravity if the point we are trying to hit is more than 2m away
	if range > 40 then
		local timeToTarget = range / muzVel
		
		-- lead the target if target speed and projectile TTT is above the threshold
		if timeToTarget * TargetVel.Magnitude > 0.5 then
			AimPoint = AimPoint + TargetVel * timeToTarget
			Dist = SceneMan:ShortestDistance(StartPos, AimPoint, false)
		end
		
		Dist = Dist / FrameMan.PPM	-- convert from pixels to meters
		local velSqr = muzVel*muzVel
		local gravity = SceneMan.GlobalAcc.Y	-- * self.Magazine.Ammo.Round.Particle.GlobalAccScalar
		local root = math.sqrt(velSqr*velSqr - gravity*(gravity*Dist.X*Dist.X+2*-Dist.Y*velSqr))
		
		if root ~= root then
			return nil	-- no solution exists if the root is NaN
		end
		
		return math.atan2(velSqr-root, gravity*Dist.X)
	end
	
	return Dist.AbsRadAngle
end

-- in sentry behavior the agent only looks for new enemies, it sometimes sharp aims to increse spotting range
function HumanBehaviors.Sentry(self)
	local sweepUp = true
	local sweepDone = false
	local maxAng = 1.4
	local minAng = -1.4
	local aim
	
	if self.OldTargetPos then	-- try to reaquire an old target
		local Dist = SceneMan:ShortestDistance(self.EyePos, self.OldTargetPos, false)
		self.OldTargetPos = nil
		if (Dist.X < 0 and self.HFlipped) or (Dist.X > 0 and not self.HFlipped) then	-- we are facing the target	
			self.deviceState = AHuman.AIMING
			self.Ctrl.AnalogAim = Dist.Normalized
			
			for i = 1, 30 do
				coroutine.yield()	-- aim here for ~0.5s
			end
		end
	elseif not self.isPlayerOwned then -- face the most likely enemy approach direction
		for i = 1, math.random(5) do	-- wait for a while
			coroutine.yield()
		end
		
		self:AddAISceneWaypoint(Vector(self.Pos.X, 0))
		self:UpdateMovePath()
		coroutine.yield()	-- wait until next frame
		
		-- face the direction of the first waypoint
		for WptPos in self.MovePath do
			local Dist = SceneMan:ShortestDistance(self.Pos, WptPos, false)
			if Dist.X > 5 then
				self.SentryFacing = false
				self.Ctrl.AnalogAim = Dist.Normalized
			elseif Dist.X < -5 then
				self.SentryFacing = true
				self.Ctrl.AnalogAim = Dist.Normalized
			end
			
			break
		end
	end
	
	while true do	-- start by looking forward
		aim = self:GetAimAngle(false)
		
		if sweepUp then
			if aim < maxAng/3 then
				self.Ctrl:SetState(Controller.AIM_UP, false)
				coroutine.yield()	-- wait until next frame
				self.Ctrl:SetState(Controller.AIM_UP, true)
			else
				sweepUp = false
			end
		else
			if aim > minAng/3 then
				self.Ctrl:SetState(Controller.AIM_DOWN, false)
				coroutine.yield()	-- wait until next frame
				self.Ctrl:SetState(Controller.AIM_DOWN, true)
			else
				sweepUp = true
				if sweepDone then
					break
				else
					sweepDone = true
				end
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	if self.HFlipped ~= self.SentryFacing then
		self.HFlipped = self.SentryFacing	-- turn to the direction we have been order to guard
		return true	-- restart this behavior
	end
	
	while true do	-- look down
		aim = self:GetAimAngle(false)
		if aim > minAng then
			self.Ctrl:SetState(Controller.AIM_DOWN, true)
		else
			break
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	local Hit = Vector()
	local NoObstacle = {}
	local StartPos
	self.deviceState = AHuman.AIMING
	
	while true do	-- scan the area for obstacles
		aim = self:GetAimAngle(false)
		if aim < maxAng then
			self.Ctrl:SetState(Controller.AIM_UP, true)
		else
			break
		end
		
		if self:EquipFirearm(false) and self.EquippedItem then
			StartPos = ToHeldDevice(self.EquippedItem).MuzzlePos
		else
			StartPos = self.EyePos
		end
		
		-- save the angle to a table if there is no obstacle
		if not SceneMan:CastStrengthRay(StartPos, Vector(60, 0):RadRotate(self:GetAimAngle(true)), 5, Hit, 2, 0, true) then
			table.insert(NoObstacle, aim)	-- TODO: don't use a table for this
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	local SharpTimer = Timer()
	local aimTime = 2000
	local angDiff = 1
	self.deviceState = AHuman.POINTING
	
	if #NoObstacle > 1 then	-- only aim where we know there are no obstacles, e.g. out of a gun port
		minAng = NoObstacle[1] * 0.95
		maxAng = NoObstacle[#NoObstacle] * 0.95
		angDiff = 1 / math.max(math.abs(maxAng - minAng), 0.1)	-- sharp aim longer from a small aiming window
	end
	
	while true do
		if not self:EquipFirearm(false) and not self:EquipThrowable(false) then
			break
		end
		
		aim = self:GetAimAngle(false)
		
		if sweepUp then
			if aim < maxAng then
				if aim < maxAng/5 and aim > minAng/5 and PosRand() > 0.3 then
					self.Ctrl:SetState(Controller.AIM_UP, false)
				else
					self.Ctrl:SetState(Controller.AIM_UP, true)
				end
			else
				sweepUp = false
			end
		else
			if aim > minAng then
				if aim < maxAng/5 and aim > minAng/5 and PosRand() > 0.3 then
					self.Ctrl:SetState(Controller.AIM_DOWN, false)
				else
					self.Ctrl:SetState(Controller.AIM_DOWN, true)
				end
			else
				sweepUp = true
			end
		end
		
		if SharpTimer:IsPastSimMS(aimTime) then
			SharpTimer:Reset()
			
			if self.deviceState == AHuman.AIMING then
				aimTime = RangeRand(1000, 3000)
				self.deviceState = AHuman.POINTING
			else
				aimTime = RangeRand(6000, 12000) * angDiff
				self.deviceState = AHuman.AIMING
			end
			
			if SceneMan:ShortestDistance(self.Pos, self.SentryPos, false).Magnitude > self.Height*0.7 then
				self.SentryPos = SceneMan:MovePointToGround(self.SentryPos, self.Height*0.25, 2)
				self:ClearAIWaypoints()
				self:AddAISceneWaypoint(self.SentryPos)
				self:CreateGoToBehavior()	-- try to return to the sentry pos
				break
			elseif self.HFlipped ~= self.SentryFacing then
				self.HFlipped = self.SentryFacing	-- turn to the direction we have been order to guard
				break	-- restart this behavior
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	return true
end


function HumanBehaviors.Patrol(self)
	while self.flying or self.Vel.Magnitude > 4 do	-- wait untill we are stationary
		return true
	end
	
	local Free = Vector()
	local WptA, WptB
	
	-- look for a path to the right
	SceneMan:CastObstacleRay(self.Pos, Vector(512, 0), Vector(), Free, self.ID, self.IgnoresWhichTeam, rte.grassID, 4)
	local Dist = SceneMan:ShortestDistance(self.Pos, Free, false)
	
	if Dist.Magnitude > 20 then
		self:ClearAIWaypoints()
		self:AddAISceneWaypoint(Free)
		coroutine.yield()	-- wait until next frame
		self:UpdateMovePath()
		coroutine.yield()	-- wait until next frame
		
		local PrevPos = Vector(self.Pos.X, self.Pos.Y)
		for WptPos in self.MovePath do
			if math.abs(PrevPos.Y - WptPos.Y) > 14 then
				break
			end
			
			WptA = Vector(PrevPos.X, PrevPos.Y)
			PrevPos:SetXY(WptPos.X, WptPos.Y)
		end
	end
	
	-- look for a path to the left
	SceneMan:CastObstacleRay(self.Pos, Vector(-512, 0), Vector(), Free, self.ID, self.IgnoresWhichTeam, rte.grassID, 4)
	Dist = SceneMan:ShortestDistance(self.Pos, Free, false)
	
	if Dist.Magnitude > 20 then
		self:ClearAIWaypoints()
		self:AddAISceneWaypoint(Free)
		coroutine.yield()	-- wait until next frame
		self:UpdateMovePath()
		coroutine.yield()	-- wait until next frame
		
		local PrevPos = Vector(self.Pos.X, self.Pos.Y)
		for WptPos in self.MovePath do
			if math.abs(PrevPos.Y - WptPos.Y) > 14 then
				break
			end
			
			WptB = Vector(PrevPos.X, PrevPos.Y)
			PrevPos:SetXY(WptPos.X, WptPos.Y)
		end
	end	
	
	self:ClearAIWaypoints()
	coroutine.yield()	-- wait until next frame
	
	if WptA then
		Dist = SceneMan:ShortestDistance(self.Pos, WptA, false)
		if Dist.Magnitude > 20 then
			self:AddAISceneWaypoint(WptA)
		else
			WptA = nil
		end
	end
	
	if WptB then
		Dist = SceneMan:ShortestDistance(self.Pos, WptB, false)
		if Dist.Magnitude > 20 then
			self:AddAISceneWaypoint(WptB)
		else
			WptB = nil
		end
	end
	
	if WptA or WptB then
		self:CreateGoToBehavior()
	else	-- no path was found
		local FlipTimer = Timer()
		FlipTimer:SetSimTimeLimitMS(3000)
		while true do
			coroutine.yield()	-- wait until next frame
			if FlipTimer:IsPastSimTimeLimit() then
				FlipTimer:Reset()
				FlipTimer:SetSimTimeLimitMS(RangeRand(2000, 5000))
				self.HFlipped = not self.HFlipped	-- turn around and try the other direction sometimes
				if PosRand() < 0.3 then
					break	-- end the behavior
				end
			end
		end
	end
	
	return true
end


function HumanBehaviors.GoldDig(self)
	if not self:EquipDiggingTool(true) then
		self:CreateGetToolBehavior()
		return true
	end
	
	local aimAngle = 0.57
	local LookVec = Vector(200,0):RadRotate(aimAngle)
	local GoldPos = Vector()
	local GoldTable = {}
	
	while true do
		self.Ctrl.AnalogAim = LookVec.Normalized
		if SceneMan:CastMaterialRay(self.EyePos, LookVec, rte.goldID, GoldPos, 2, true) then
			coroutine.yield()	-- wait until next frame
			
			local str = SceneMan:CastStrengthSumRay(self.EyePos, self.EyePos+LookVec, 4, rte.goldID)/5	-- prioritise gold in soft ground
			local Dist = SceneMan:ShortestDistance(self.Pos, GoldPos, false)	-- prioritise gold close to us
			
			if GoldPos.Y > SceneMan.SceneHeight - 40 then
				str = str + 3000	-- avoid gold close to the lower edege of the scene
			end
			
			if not SceneMan.SceneWrapsX and GoldPos.X < 40 or GoldPos.X > SceneMan.SceneWidth - 40 then
				str = str + 2000	-- avoid gold close to the edeges of a non-wrapping scene
			end
			
			if Dist.Magnitude < 30 then	-- dig to a point behind the gold
				GoldPos = self.Pos + Dist:SetMagnitude(40)
			end
			
			table.insert(GoldTable, {Pos=Vector(GoldPos.X, GoldPos.Y), prio=str+Dist.Magnitude})
		end
		
		if aimAngle < -3.71 then
			break
		else
			aimAngle = aimAngle - 0.025
			LookVec = Vector(200,0):RadRotate(aimAngle)
			coroutine.yield()	-- wait until next frame
		end
	end
	
	if #GoldTable > 0 then
		table.sort(GoldTable, function(A, B) return A.prio < B.prio end)
		coroutine.yield()	-- wait until next frame
		
		self:ClearAIWaypoints()
		self:AddAISceneWaypoint(GoldTable[1].Pos)
		self:CreateGoToBehavior()
	elseif self.Pos.Y + 200 < SceneMan.SceneHeight then	-- don't dig beyond the scene limit
		-- no gold found, so dig down and try again
		local str_r = SceneMan:CastStrengthSumRay(self.Pos, self.Pos+Vector( 100, 100), 5, rte.goldID)
		coroutine.yield()	-- wait until next frame
		local str_l = SceneMan:CastStrengthSumRay(self.Pos, self.Pos+Vector(-100, 100), 5, rte.goldID)
		coroutine.yield()	-- wait until next frame
		
		self:ClearAIWaypoints()
		if str_r < str_l then
			self:AddAISceneWaypoint(self.Pos+Vector( 100, 100))
		else
			self:AddAISceneWaypoint(self.Pos+Vector(-100, 100))
		end
		
		self:CreateGoToBehavior()
	end
	
	return true
end


-- find the closest enemy brain
function HumanBehaviors.BrainSearch(self)
	local Activ = ActivityMan:GetActivity()
	local Brains = {}
	
	for Act in MovableMan.Actors do
		if Act.Team ~= self.Team and Act:HasObjectInGroup("Brains") then
			table.insert(Brains, Act)
		end
	end
	
	coroutine.yield()	-- wait until next frame
	
	if #Brains < 1 then	-- no bain actors found, check if some other actor is the brain
		for player = Activity.PLAYER_1, Activity.MAXPLAYERCOUNT - 1 do
			if Activ:PlayerActive(player) and Activ:GetTeamOfPlayer(player) ~= self.Team then
				local Target = Activ:GetPlayerBrain(player)
				if Target and MovableMan:IsActor(Target) then
					table.insert(Brains, Target)
				end
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	if #Brains > 0 then
		if #Brains == 1 then
			if MovableMan:IsActor(Brains[1]) then
				self:ClearAIWaypoints()
				self.FollowingActor = Brains[1]
				self:AddAIMOWaypoint(self.FollowingActor)
				self:CreateGoToBehavior()
			end
		else
			local ClosestBrain
			local minDist = math.huge
			for _, Act in pairs(Brains) do
				if MovableMan:IsActor(Act) then
					self:ClearAIWaypoints()
					self:AddAISceneWaypoint(Act.Pos)
					self:UpdateMovePath()
					
					local waypoints = self.MovePathSize
					if waypoints < minDist then
						minDist = waypoints
						ClosestBrain = Act
					end
					
					coroutine.yield()	-- wait until next frame
				end
			end
			
			self:ClearAIWaypoints()
			
			if MovableMan:IsActor(ClosestBrain) then
				self.FollowingActor = ClosestBrain
				self:AddAIMOWaypoint(self.FollowingActor)
				self:CreateGoToBehavior()
			else
				return true	-- the brain we found died while we where searching, restart this behavior next frame
			end
		end
	else
		self.AIMode = Actor.AIMODE_PATROL	-- no enemy brains left
	end
	
	return true
end


-- find a weapon to pick up
function HumanBehaviors.WeaponSearch(self)
	local range, minDist, HD
	local Devices = {}
	local pickupDiggers = not self:EquipDiggingTool(false)
	
	if self.isPlayerOwned then
		minDist = 70	-- don't move player actors more than 3.5m
	else
		minDist = FrameMan.PlayerScreenWidth * 0.45
	end
	
	for Item in MovableMan.Items do	-- store all HeldDevices of the correct type and within a cerain range in a table
		HD = ToHeldDevice(Item)
		if not HD:IsActivated() and HD.Vel.Magnitude < 2 then
			range = SceneMan:ShortestDistance(self.Pos, HD.Pos, false).Magnitude
			if range < minDist and not SceneMan:IsUnseen(HD.Pos.X, HD.Pos.Y, self.Team) then
				table.insert(Devices, HD)
			end
		end
	end
	
	if #Devices > 0 then
		coroutine.yield()	-- wait until next frame
		
		local PrevWpt
		if not self.FollowingActor then
			if self.AIMode == Actor.AIMODE_SENTRY and not self.flying then
				PrevWpt = Vector(self.Pos.X, self.Pos.Y)	-- return here after pick up
			else
				PrevWpt = self:GetLastAIWaypoint()
				range = SceneMan:ShortestDistance(self.Pos, PrevWpt, false).Magnitude
				if range < 1 then
					PrevWpt = nil	-- not a valid waypoint
				end
			end
		end
		
		if self.isPlayerOwned then
			minDist = 10	-- # of waypoints
		else
			minDist = 48
		end
		
		local waypoints
		local DevicesToPickUp = {}
		for _, Item in pairs(Devices) do
			if MovableMan:IsDevice(Item) then
				self:ClearAIWaypoints()
				self:AddAISceneWaypoint(Item.Pos)
				self:UpdateMovePath()
				
				-- estimate the walking distance to the item				
				if Item.ClassName == "TDExplosive" then
					waypoints = self.MovePathSize * 1.4	-- prioritize non-grenades
				elseif Item:IsTool() then
					if pickupDiggers then
						waypoints = self.MovePathSize * 1.8	-- prioritize non-diggers
					else
						waypoints = minDist -- don't pick up
					end
				else
					waypoints = self.MovePathSize
				end
				
				if waypoints < minDist then
					table.insert(DevicesToPickUp, {HD=Item, range=waypoints})
				end
				
				coroutine.yield()	-- wait until next frame
			end
		end
		
		self:ClearAIWaypoints()
		table.sort(DevicesToPickUp, function(A,B) return A.range < B.range end)	-- sort the items in order of distance
		coroutine.yield()	-- wait until next frame
		
		self.PickupHD = nil
		for _, Data in pairs(DevicesToPickUp) do
			if MovableMan:IsDevice(Data.HD) then
				self.PickupHD = Data.HD
				break
			end
		end
		
		if self.PickupHD then
			if PrevWpt then
				self.PrevAIWaypoint = PrevWpt
			end
			
			self:AddAIMOWaypoint(self.PickupHD)
			self:CreateGoToBehavior()
		else	-- no items found
			if self.FollowingActor and MovableMan:IsActor(self.FollowingActor) then
				self:AddAIMOWaypoint(self.FollowingActor)
			elseif self.PrevAIWaypoint then
				self:AddAISceneWaypoint(self.PrevAIWaypoint)
			end
		end
	end
	
	return true
end


-- find a tool to pick up
function HumanBehaviors.ToolSearch(self)
	local range, minDist, HD
	local Devices = {}
	
	if self.AIMode == Actor.AIMODE_GOLDDIG then
		minDist = FrameMan.PlayerScreenWidth * 0.5	-- move up to half a screen when digging
	elseif self.isPlayerOwned then
		minDist = 50	-- don't move player actors more than 2.5m
	else
		minDist = FrameMan.PlayerScreenWidth * 0.3
	end
	
	for Item in MovableMan.Items do	-- store all HeldDevices of the correct type and within a cerain range in a table
		HD = ToHeldDevice(Item)
		if HD and HD:IsTool() and HD.Vel.Magnitude < 2 then
			range = SceneMan:ShortestDistance(self.Pos, HD.Pos, false).Magnitude
			if range < minDist and not SceneMan:IsUnseen(HD.Pos.X, HD.Pos.Y, self.Team) then
				table.insert(Devices, HD)
			end
		end
	end
	
	if #Devices > 0 then
		coroutine.yield()	-- wait until next frame
		
		local PrevWpt
		if not self.FollowingActor then
			if self.AIMode == Actor.AIMODE_SENTRY and not self.flying then
				PrevWpt = Vector(self.Pos.X, self.Pos.Y)	-- return here after pick up
			else
				PrevWpt = self:GetLastAIWaypoint()
				range = SceneMan:ShortestDistance(self.Pos, PrevWpt, false).Magnitude
				if range < 1 then
					PrevWpt = nil	-- not a valid waypoint
				end
			end
		end
		
		if self.AIMode == Actor.AIMODE_GOLDDIG then
			minDist = 30
		elseif self.isPlayerOwned then
			minDist = 5
		else
			minDist = 15
		end
		
		for _, Item in pairs(Devices) do
			if MovableMan:IsDevice(Item) then
				self:ClearAIWaypoints()
				self:AddAISceneWaypoint(Item.Pos)
				self:UpdateMovePath()
				
				local waypoints = self.MovePathSize	-- estimate the walking distance to the item
				if waypoints < minDist then
					minDist = waypoints
					self.PickupHD = Item
				end
				
				coroutine.yield()	-- wait until next frame
			end
		end
		
		self:ClearAIWaypoints()
		
		if MovableMan:IsDevice(self.PickupHD) then		
			if PrevWpt then
				self.PrevAIWaypoint = PrevWpt
			end
			
			self:AddAIMOWaypoint(self.PickupHD)
			self:CreateGoToBehavior()
		else
			self.PickupHD = nil	-- the item became invalid while searching
			
			if self.FollowingActor and MovableMan:IsActor(self.FollowingActor) then
				self:AddAIMOWaypoint(self.FollowingActor)
			elseif self.PrevAIWaypoint then
				self:AddAISceneWaypoint(self.PrevAIWaypoint)
			end
		end
	end
	
	return true
end


-- move to the next waypoint
function HumanBehaviors.GoToWpt(self)
	if not self.MOMoveTarget then
		if SceneMan:ShortestDistance(self:GetLastAIWaypoint(), self.Pos, false).Magnitude < 10 then
			self:ClearAIWaypoints()
			if self.AIMode == Actor.AIMODE_GOTO then
				self.AIMode = Actor.AIMODE_SENTRY
			end
			return true
		end
	end
	
	-- is Y1 lower down in the scene, compared to Y2?
	local Lower = function(Y1, Y2, margin)
		return Y1.Pos.Y - margin > Y2.Pos.Y
	end
	
	-- is Y1 higher up in the scene, compared to Y2?
	local Higher = function(Y1, Y2, margin)
		return Y1.Pos.Y + margin < Y2.Pos.Y
	end
	
	local AngleBetweenVectors = function(A, B)	-- returns the smallest angle between two vectors
		local angle = math.atan2(A.Y, A.X) - math.atan2(B.Y, B.X)
		
		if angle > 3.142 then
		 angle = angle - 6.283
		elseif angle < -3.142 then
		 angle = angle + 6.283
		end
		
		return angle
	end
	
	local UpdatePathTimer = Timer()
	local ArrivedTimer = Timer()
	local BurstTimer = Timer()
	
	local NoLOSTimer = Timer()
	NoLOSTimer:SetSimTimeLimitMS(1000)
	
	local StuckTimer = Timer()
	StuckTimer:SetSimTimeLimitMS(2000)
	
	local nextLatMove = self.lateralMoveState
	local nextAimAngle = self:GetAimAngle(false) * 0.8
	local scanAng = 0	-- for obstacle detection
	local Obstacles = {}
	local PrevWptPos = self.Pos
	local sweepCW = true
	local sweepRange = 0
	local digState = AHuman.NOTDIGGING
	local obstacleState = Actor.PROCEEDING
	local WptList, Waypoint, Dist, CurrDist
	
	while true do
		while not self.flying and self.Target and (self:EquipFirearm(false) or self:EquipThrowable(false)) do	-- don't move around if we have something to shoot at
			coroutine.yield()	-- wait until next frame
		end
		
		if self.Vel.Magnitude > 2 then
			StuckTimer:Reset()
		end
		
		if self.MOMoveTarget then	-- make the last waypoint marker stick to the MO we are following
			if MovableMan:ValidMO(self.MOMoveTarget) then
				self:RemoveMovePathEnd()
				self:AddToMovePathEnd(self.MOMoveTarget.Pos)
				
				if self.Team == self.MOMoveTarget.Team and not self.flying then	-- we are following an ally, stop when close
					while SceneMan:ShortestDistance(self.Pos, self.MOMoveTarget.Pos, false).Magnitude < self.Radius+self.MOMoveTarget.Radius+10 do
						coroutine.yield()	-- wait until next frame
						
						if not MovableMan:ValidMO(self.MOMoveTarget) then
							self:ClearMovePath()
							break
						end
					end
				end
			else
				self:ClearMovePath()
			end
		end
		
		if not self.flying and UpdatePathTimer:IsPastSimMS(6000) then
			if Waypoint and self.BlockingActor then
				if MovableMan:IsActor(self.BlockingActor) then
					CurrDist = SceneMan:ShortestDistance(self.Pos, Waypoint.Pos, false)
					if (self.Pos.X > self.BlockingActor.Pos.X and CurrDist.X < self.Pos.X) or
						(self.Pos.X < self.BlockingActor.Pos.X and CurrDist.X > self.Pos.X) or
						SceneMan:ShortestDistance(self.Pos, self.BlockingActor.Pos, false).Magnitude > self.Diameter + self.BlockingActor.Diameter
					then
						self.BlockingActor = nil	-- the blocking actor is not in the way any longer
						self.teamBlockState = Actor.NOTBLOCKED
					else
						self.BlockedTimer:Reset()
						self.teamBlockState = Actor.IGNORINGBLOCK
						self:CreateMoveAroundBehavior()
						break	-- end this behavior
					end
				else
					self.BlockingActor = nil
				end
			end
			
			UpdatePathTimer:Reset()
			self.deviceState = AHuman.STILL
			self.crawling = false
			self.jump = false
			nextLatMove = Actor.LAT_STILL
			Waypoint = nil
			WptList = nil
		elseif StuckTimer:IsPastSimTimeLimit() then	-- dislodge
			if self.jump then
				if self.JetTimeLeft < TimerMan.DeltaTimeMS * 3 then	-- out of fuel
					self.jump = false
					self.refuel = true
					nextLatMove = Actor.LAT_STILL
				else
					local chance = PosRand()
					if chance < 0.1 then
						nextLatMove = Actor.LAT_LEFT
					elseif chance > 0.9 then
						nextLatMove = Actor.LAT_RIGHT
					else
						nextLatMove = Actor.LAT_STILL
					end
				end
			else
				if PosRand() < 0.2 then
					if self.lateralMoveState == Actor.LAT_LEFT then
						nextLatMove = Actor.LAT_RIGHT
					else
						nextLatMove = Actor.LAT_LEFT
					end
				end
				
				if self.refuel and self.JetTimeLeft > self.JetTimeTotal * 0.95 then
					self.jump = true
					BurstTimer:SetSimTimeLimitMS(500)	-- this burst last until the BurstTimer expire
					BurstTimer:Reset()
				end
			end
		elseif WptList then	-- we have a list of waypoints, folow it
			if #WptList < 1 and not Waypoint then	-- arrived
				if self.MOMoveTarget and MovableMan:ValidMO(self.MOMoveTarget) then
					if self.Team ~= self.MOMoveTarget.Team then
						Waypoint = {Pos=self.MOMoveTarget.Pos}
					else
						nextLatMove = Actor.LAT_STILL
						self.jump = false
						coroutine.yield()	-- wait until next frame
					end
				else
					break
				end
			else
				if not Waypoint then	-- get the next waypoint in the list					
					-- we must update the path more often if we are following someone
					if self.MOMoveTarget and MovableMan:ValidMO(self.MOMoveTarget) then
						if SceneMan:ShortestDistance(self:GetLastAIWaypoint(), self.MOMoveTarget.Pos, false).Magnitude > 100 then	-- the target has moved
							WptList = nil
							Waypoint = nil
						else
							UpdatePathTimer:Reset()
						end
					end
					
					if WptList then						
						local NextWptPos = WptList[1].Pos
						Dist = SceneMan:ShortestDistance(self.Pos, NextWptPos, false)
						if Dist.Y < -25 and math.abs(Dist.X) < 30 then	-- avoid any corners if the next waypoint is above us
							local CornerPos = Vector(NextWptPos.X, NextWptPos.Y)
							if self.Pos.X > CornerPos.X then
								CornerPos = CornerPos + Vector(25, -40)
								cornerType = "right"
							else
								CornerPos = CornerPos + Vector(-25, -40)
								cornerType = "left"
							end
							
							local Free = Vector()
							Dist = SceneMan:ShortestDistance(NextWptPos, CornerPos, false)
							-- make sure the corner waypoint is not inside terrain
							local pixels = SceneMan:CastObstacleRay(NextWptPos, Dist, Vector(), Free, self.ID, self.IgnoresWhichTeam, rte.grassID, 3)
							if pixels == 0 then
								break	-- the waypoint is inside terrain, plot a new path
							elseif pixels > 0 then
								CornerPos = (NextWptPos + Free) / 2	-- compensate for obstacles
							end
							
							coroutine.yield()	-- wait until next frame
							
							-- check if we have LOS
							Dist = SceneMan:ShortestDistance(self.Pos, CornerPos, false)
							if 0 <= SceneMan:CastObstacleRay(self.Pos, Dist, Vector(), Vector(), self.ID, self.IgnoresWhichTeam, rte.grassID, 4) then
								-- CornerPos is blocked
								CornerPos.X = self.Pos.X	-- move CornerPos straight above us
								cornerType = "air"
							end
							
							coroutine.yield()	-- wait until next frame
							
							Waypoint = {Pos=CornerPos, Type=cornerType}
							if #WptList > 1 and not WptList[1].Type then	-- remove the waypoint after the corner if possible
								table.remove(WptList, 1)
								self:RemoveMovePathBeginning() -- clean up the graphical representation of the path
							end
							
							self:AddToMovePathBeginning(Waypoint.Pos)
						else
							Waypoint = table.remove(WptList, 1)							
							if Waypoint.Type ~= "air" then
								local Free = Vector()
								
								-- only if we have a digging tool
								if Waypoint.Type ~= "drop" and self:EquipDiggingTool(false) then
									local PathSegRay = SceneMan:ShortestDistance(PrevWptPos, Waypoint.Pos, false)	-- detect material blocking the path and start digging through it
									if self.teamBlockState ~= Actor.BLOCKED and SceneMan:CastStrengthRay(PrevWptPos, PathSegRay, 5, Free, 1, rte.doorID, true) then
										if SceneMan:ShortestDistance(self.Pos, Free, false).Magnitude < self.Height*0.5 then	-- check that we're close enough to start digging
											-- TODO: update the path and try again just to make sure we can't get through some where else
											digState = AHuman.STARTDIG
											self.deviceState = AHuman.DIGGING
											obstacleState = Actor.DIGPAUSING
											nextLatMove = Actor.LAT_STILL
											sweepRange = math.pi / 4
											StuckTimer:SetSimTimeLimitMS(5000)
											self.Ctrl.AnalogAim = SceneMan:ShortestDistance(self.Pos, Waypoint.Pos, false).Normalized	-- aim in the direction of the next waypoint
										else
											digState = AHuman.NOTDIGGING
											obstacleState = Actor.PROCEEDING
										end
										
										coroutine.yield()	-- wait until next frame
									else
										digState = AHuman.NOTDIGGING
										obstacleState = Actor.PROCEEDING
										StuckTimer:SetSimTimeLimitMS(1500)
									end
								end
								
								if digState == AHuman.NOTDIGGING and self.deviceState ~= AHuman.DIGGING then
									-- if our path isn't blocked enough to dig, but the headroom is too little, start crawling to get through
									local Heading = SceneMan:ShortestDistance(self.Pos, Waypoint.Pos, false):SetMagnitude(self.Height*0.5)
									
									-- don't crawl if it's too steep, climb then instead
									if math.abs(Heading.X) > math.abs(Heading.Y) and self.Head and self.Head:IsAttached() then
										local TopHeadPos = self.Head.Pos - Vector(0, self.Head.Radius*0.7)
										
										-- first check up to the top of the head, and then from there forward
										if SceneMan:CastStrengthRay(self.Pos, TopHeadPos - self.Pos, 5, Free, 4, rte.doorID, true) or
												SceneMan:CastStrengthRay(TopHeadPos, Heading, 5, Free, 4, rte.doorID, true)
										then
											self.crawling = true
										else
											self.crawling = false
										end
										
										coroutine.yield()	-- wait until next frame
									else
										self.crawling = false
									end
								end
							end
						end
						
						if not Waypoint.Type then
							ArrivedTimer:SetSimTimeLimitMS(100)
						elseif Waypoint.Type == "last" then
							ArrivedTimer:SetSimTimeLimitMS(500)
						else	-- air or corner wpt
							ArrivedTimer:SetSimTimeLimitMS(25)
						end
					end
				elseif #WptList > 1 then	-- check if some other waypoint is closer
					local test = math.random(1, #WptList)
					local RandomWpt = WptList[test]
					if RandomWpt then
						Dist = SceneMan:ShortestDistance(self.Pos, RandomWpt.Pos, false)
						local mag = Dist.Magnitude
						if mag < 50 and mag < SceneMan:ShortestDistance(self.Pos, Waypoint.Pos, false).Magnitude/3 then
							-- this waypoint is closer, check LOS
							if -1 == SceneMan:CastObstacleRay(self.Pos, Dist, Vector(), Vector(), self.ID, self.IgnoresWhichTeam, rte.grassID, 4) then
								Waypoint = RandomWpt	-- go here instead
								if WptList[test-1] then
									PrevWptPos = WptList[test-1].Pos
								else
									PrevWptPos = self.Pos
								end
								
								test = math.min(test, #WptList)
								for i = 1, test do	-- delete the earlier waypoints
									table.remove(WptList, 1)
									if #WptList > 0 then
										self:RemoveMovePathBeginning()
									end
								end
							end
						end
					end
				end
				
				if Waypoint then
					CurrDist = SceneMan:ShortestDistance(self.Pos, Waypoint.Pos, false)
					
					-- digging
					if digState ~= AHuman.NOTDIGGING then
						if not self.Target and self:EquipDiggingTool(true) then	-- switch to the digger if we have one
							if ToAHuman(self).FirearmIsEmpty then	-- reload if it's empty
								self.fire = false
								self.Ctrl:SetState(Controller.WEAPON_RELOAD, true)
							else
								if self.teamBlockState == Actor.BLOCKED then
									self.fire = false
									nextLatMove = Actor.LAT_STILL
								else
									if obstacleState == Actor.PROCEEDING then
										if CurrDist.X < -1 then
											nextLatMove = Actor.LAT_LEFT
										elseif CurrDist.X > 1 then
											nextLatMove = Actor.LAT_RIGHT
										end
									else
										nextLatMove = Actor.LAT_STILL
									end
									
									-- check if we are close enough to dig
									if SceneMan:ShortestDistance(PrevWptPos, self.Pos, false).Magnitude > self.Height*0.5 and
										 SceneMan:ShortestDistance(self.Pos, Waypoint.Pos, false).Magnitude > self.Height*0.5
									then
										digState = AHuman.NOTDIGGING
										obstacleState = Actor.PROCEEDING
										self.deviceState = AHuman.STILL
										self.fire = false
										self:EquipFirearm(true)
									else
										local aimAngle = self:GetAimAngle(true)
										local AimVec = Vector(1, 0):RadRotate(aimAngle)
										
										-- see if we have dug out all that we can in the sweep area without moving closer
										local centerAngle = CurrDist.AbsRadAngle
										local Ray = Vector(self.Height*0.3, 0):RadRotate(centerAngle)	-- center
										if SceneMan:CastNotMaterialRay(self.Pos, Ray, 0, 3, false) < 0 then
											coroutine.yield()	-- wait until next frame
											
											-- now check the tunnel's thickness
											Ray = Vector(self.Height*0.3, 0):RadRotate(centerAngle + sweepRange)	-- up
											if SceneMan:CastNotMaterialRay(self.Pos, Ray, rte.airID, 3, false) < 0 then
												coroutine.yield()	-- wait until next frame
												
												Ray = Vector(self.Height*0.3, 0):RadRotate(centerAngle - sweepRange)	-- down
												if SceneMan:CastNotMaterialRay(self.Pos, Ray, rte.airID, 3, false) < 0 then
													obstacleState = Actor.PROCEEDING	-- ok the tunnel section is clear, so start walking forward while still digging
												else
													obstacleState = Actor.DIGPAUSING	-- tunnel cavity not clear yet, so stay put and dig some more
												end
											end
										else
											obstacleState = Actor.DIGPAUSING	-- tunnel cavity not clear yet, so stay put and dig some more
										end
										
										coroutine.yield()	-- wait until next frame
										
										local angDiff = AngleBetweenVectors(CurrDist, AimVec)
										if math.abs(angDiff) < sweepRange then
											self.fire = true	-- only fire the digger at the obstacle
										else
											self.fire = false
										end
										
										-- sweep the digger between the two endpoints of the obstacle
										local DigTarget
										if sweepCW then
											DigTarget = Vector(self.Height*0.4, 0):RadRotate(centerAngle + sweepRange)
										else
											DigTarget = Vector(self.Height*0.4, 0):RadRotate(centerAngle - sweepRange)
										end
										
										angDiff = AngleBetweenVectors(DigTarget, AimVec)
										if math.abs(angDiff) < 0.1 then
											sweepCW = not sweepCW	-- this is close enough, go in the other direction next frame
										else
											self.Ctrl.AnalogAim = Vector(AimVec.X, AimVec.Y):RadRotate(-angDiff*0.1)
										end
										
										-- check if we are done when we get close enough to the waypoint
										if SceneMan:ShortestDistance(self.Pos, Waypoint.Pos, false).Magnitude < self.Height*0.25 then
											if not SceneMan:CastStrengthRay(PrevWptPos, SceneMan:ShortestDistance(PrevWptPos, Waypoint.Pos, false), 5, Vector(), 1, rte.doorID, true) and
												 not SceneMan:CastStrengthRay(self.EyePos, SceneMan:ShortestDistance(self.EyePos, Waypoint.Pos, false), 5, Vector(), 1, rte.doorID, true)
											then
												-- advance to the next waypoint, if there are any
												if #WptList > 0 then
													UpdatePathTimer:Reset()
													PrevWptPos = Waypoint.Pos
													Waypoint = table.remove(WptList, 1)
													if #WptList > 0 then
														self:RemoveMovePathBeginning()
													end
												end
											end
											
											coroutine.yield()	-- wait until next frame
										end
									end
								end
							end
						else
							digState = AHuman.NOTDIGGING
							obstacleState = Actor.PROCEEDING
							self.deviceState = AHuman.STILL
							self.fire = false
							self:EquipFirearm(true)
						end
					else	-- not digging
						if not self.Target then
							self.fire = false
						end
						
						-- Scan for obstacles
						local Trace = Vector(self.Diameter*0.85, 0):RadRotate(scanAng)
						local Free = Vector()
						local index = math.floor(scanAng*2.5+2.01)
						if SceneMan:CastObstacleRay(self.Pos, Trace, Vector(), Free, self.ID, self.IgnoresWhichTeam, rte.grassID, 3) > -1 then
							Obstacles[index] = true
						else
							Obstacles[index] = false
						end
						
						if scanAng < 1.57 then	-- pi/2
							if scanAng > 1.2 then
								scanAng = 1.89
							else
								scanAng = scanAng + 0.55
							end
						else
							if scanAng > 3.5 then
								scanAng = -0.4
							else
								scanAng = scanAng + 0.55
							end
						end
						
						coroutine.yield()	-- wait until next frame
						
						if CurrDist.Magnitude > self.Height/3 then	-- not close enough to the waypoint
							ArrivedTimer:Reset()
							
							-- check if we have LOS to the waypoint
							if SceneMan:CastObstacleRay(self.Pos, CurrDist, Vector(), Vector(), self.ID, self.IgnoresWhichTeam, rte.grassID, 9) < 0 then
								NoLOSTimer:Reset()
							elseif NoLOSTimer:IsPastSimTimeLimit() then	-- calulate new path
								Waypoint = nil
								WptList = nil
							end							
							
							if not self.jump and not self.flying then
								coroutine.yield()	-- wait until next frame
							end
						elseif ArrivedTimer:IsPastSimTimeLimit() then	-- only remove a waypoint if we have been close to it for a while
							PrevWptPos = Waypoint.Pos
							self:RemoveMovePathBeginning()
							Waypoint = nil
						elseif self.refuel and self.Vel.Y < 5 then
							if self.JetTimeLeft > self.JetTimeTotal * 0.95 or
								(self.flying and self.Vel.Y < -3 and self.JetTimeLeft > 0.5)
							then
								self.refuel = false
							else
								self.jump = false	-- wait until the jetpack is full
								nextLatMove = Actor.LAT_STILL
							end
						end
						
						if Waypoint and not self.refuel then	-- move towards the waypoint
							nextAimAngle = self:GetAimAngle(false) * 0.8	-- look straight ahead
							
							-- control vertical movement
							local change
							if self.flying then
								change = self.YposPID:Update(CurrDist.Y+self.Vel.Y*3, 0)
							else
								change = self.YposPID:Update(CurrDist.Y, 0) -- ignore our velocity if on the ground
							end
							
							if math.abs(self.RotAngle) < 0.5 and self.Vel.Y > -10 and
								(change < -1 or self.Vel.Y > 10)
							then
								if not self.jump then
									self.jump = true
									BurstTimer:Reset()	-- this burst last until the BurstTimer expire
								end
								
								if change < -6 then
									BurstTimer:SetSimTimeLimitMS(250)
								elseif change < -3 then
									BurstTimer:SetSimTimeLimitMS(500)
								else
									BurstTimer:SetSimTimeLimitMS(1000)
								end
							else
								self.jump = false
							end
							
							-- control horisontal movement
							if self.jump then
								if self.flying then
									change = self.XposPID:Update(CurrDist.X+self.Vel.X*3, 0)
								else
									change = self.XposPID:Update(CurrDist.X, 0) -- ignore our velocity if on the ground
								end
								
								if change < -0.5 then
									nextLatMove = Actor.LAT_LEFT
								elseif change > 0.5 then
									nextLatMove = Actor.LAT_RIGHT
								else
									nextAimAngle = 1.2	-- look up to aim jetpack down
								end
							elseif self.FGLeg or self.BGLeg then
								if CurrDist.X < -3 then
									nextLatMove = Actor.LAT_LEFT
								elseif CurrDist.X > 3 then
									nextLatMove = Actor.LAT_RIGHT
								else
									nextLatMove = Actor.LAT_STILL
								end
							elseif (CurrDist.X < -5 and self.HFlipped) or (CurrDist.X > 5 and not self.HFlipped) then
								-- no legs, jump forward
								BurstTimer:Reset()
								self.jump = true
							end
							
							if Waypoint.Type == "right" then
								if CurrDist.X > -3 then
									nextLatMove = Actor.LAT_RIGHT
								end
							elseif Waypoint.Type == "left" then
								if CurrDist.X < 3 then
									nextLatMove = Actor.LAT_LEFT
								end								
							end
							
							-- must fire thrusters to move sideways when in the air
							if not self.jump and self.flying and Waypoint.Type ~= "drop" and
								 self.Vel.Y > -5 and CurrDist.Y-self.Vel.Y*3 < 20	--TODO: use Ychange here?
							then
								change = math.abs(self.XposPID:Update(CurrDist.X+self.Vel.X*3, 0))
								if change > 0.5 then
									self.jump = true
								end
							end
							
							if self.jump then
								-- obstacle right
								for i = rte.Obst.R_FRONT, rte.Obst.R_UP do
									if Obstacles[i] then
										if Obstacles[rte.Obst.R_UP] then
											self.jump = false
										else
											nextAimAngle = 1.2	-- look up
										end
										
										break
									end
								end
								
								-- obstacle left
								for i = rte.Obst.L_UP, rte.Obst.L_FRONT do
									if Obstacles[i] then
										if Obstacles[rte.Obst.L_UP] then
											self.jump = false
										else
											nextAimAngle = 1.2	-- look up
										end
										
										break
									end
								end
							elseif Waypoint.Type ~= "drop" and not Lower(Waypoint, self, 20) then						
								-- jump over low obstacles unless we want to jump off a ledge
								if nextLatMove == Actor.LAT_RIGHT and
									(Obstacles[rte.Obst.R_LOW] or Obstacles[rte.Obst.R_FRONT]) and not Obstacles[rte.Obst.R_UP]
								then
									self.jump = true
									if Obstacles[rte.Obst.R_HIGH] then
										nextLatMove = Actor.LAT_LEFT -- TODO: only when too close to the obstacle?
									end
								elseif nextLatMove == Actor.LAT_LEFT and
									(Obstacles[rte.Obst.L_LOW] or Obstacles[rte.Obst.L_FRONT]) and not Obstacles[rte.Obst.L_UP]
								then
									self.jump = true
									if Obstacles[rte.Obst.L_HIGH] then
										nextLatMove = Actor.LAT_RIGHT -- TODO: only when too close to the obstacle?
									end
								end
							end
							
							if math.abs(self.RotAngle) > 0.5 then
								self.jump = false
							end
						end
						
						if not self.Target then
							self:SetAimAngle(nextAimAngle)
						end
					end
				end
			end
		else	-- no waypoint list, create one in several small steps to reduce lag
			self:DrawWaypoints(false)
			self:UpdateMovePath()
			coroutine.yield()	-- wait until next frame
			
			if self.MovePathSize < 1 then
				break
			end
			
			-- copy the MovePath to a temporary table so we can yield safely while working on the path
			local PathDump = {}
			for WptPos in self.MovePath do
				table.insert(PathDump, WptPos)
			end
			
			coroutine.yield()	-- wait until next frame
			
			-- copy useful waypoints to a temporary path
			local TmpWpts = {}
			table.insert(TmpWpts, {Pos=self.Pos})
			local Origin
			local LastPos = PathDump[1]
			for _, WptPos in pairs(PathDump) do
				Origin = TmpWpts[#TmpWpts].Pos
				WptPos = SceneMan:MovePointToGround(WptPos, self.Height/4, 4)
				if SceneMan:ShortestDistance(Origin, WptPos, false).Magnitude > 100 or	-- skip any waypoint too close to the previous one
					SceneMan:CastStrengthSumRay(Origin, WptPos, 3, rte.grassID) > 5
				then
					table.insert(TmpWpts, {Pos=LastPos})
				end
				
				LastPos = WptPos
				coroutine.yield()	-- wait until next frame
			end
			
			table.insert(TmpWpts, {Pos=PathDump[#PathDump]})	-- add the last waypoint in the MovePath
			
			coroutine.yield()	-- wait until next frame
			
			WptList = {}	-- store the waypoints we want in our path here
			local StartWpt = table.remove(TmpWpts, 1)
			while #TmpWpts > 0 do
				local NextWpt = table.remove(TmpWpts, 1)
				
				if Lower(NextWpt, StartWpt, 30) then	-- scan for sharp drops	TODO: check the slope instead?
					NextWpt.Type = "drop"
					
					local GapList = {}
					for j, JumpWpt in pairs(TmpWpts) do	-- look for the other side
						local Gap = SceneMan:ShortestDistance(StartWpt.Pos, JumpWpt.Pos, false)
						if Gap.Magnitude > 400 - Gap.Y then	-- TODO: use actor properties here
							break	-- too far
						end
						
						if Gap.Y > -40 then	-- no more than 2m above
							table.insert(GapList, {Wpt=JumpWpt, score=math.abs(Gap.X/Gap.Y), index=j})
						end
					end
					
					coroutine.yield()	-- wait until next frame
					
					table.sort(GapList, function(A, B) return A.score > B.score end)	-- sort largest first
					
					for _, LZ in pairs(GapList) do
						-- check if we can jump
						local Trace = SceneMan:ShortestDistance(StartWpt.Pos, LZ.Wpt.Pos, false)
						if -1 == SceneMan:CastObstacleRay(StartWpt.Pos, Trace, Vector(), Vector(), self.ID, self.IgnoresWhichTeam, rte.grassID, 4) then							
							-- find a point mid-air
							local TestPos = StartWpt.Pos + Trace / 2
							local Free = Vector()
							if 0 ~= SceneMan:CastObstacleRay(TestPos, Vector(0, -math.abs(Trace.X)/2), Vector(), Free, self.ID, self.IgnoresWhichTeam, rte.grassID, 2) then	-- TODO: check LOS? what if 0?
								table.insert(WptList, {Pos=Free+Vector(0,self.Height/4), Type="air"})	-- guide point in the air
								NextWpt = LZ.Wpt
								
								-- delete any waypoints between StartWpt and the LZ
								for i = LZ.index, 1, -1 do
									table.remove(TmpWpts, i)
								end
								
								break
							end
						end
						
						coroutine.yield()	-- wait until next frame
					end
				end
				
				table.insert(WptList, NextWpt)
				StartWpt = NextWpt
			end
			
			WptList[#WptList].Type = "last"
			
			coroutine.yield()	-- wait until next frame
			
			-- create the move path seen on the screen
			self:ClearMovePath()
			for _, Wpt in pairs(WptList) do
				self:AddToMovePathEnd(Wpt.Pos)
			end
			
			self:DrawWaypoints(true)
			NoLOSTimer:Reset()
		end
		
		-- movement commands
		if self.Target and (self:EquipFirearm(false) or self:EquipThrowable(false)) then	-- don't issue body commands if we have a target
			self.lateralMoveState = Actor.LAT_STILL
		else
			self.lateralMoveState = nextLatMove
		end
		
		if self.jump and BurstTimer:IsPastSimTimeLimit() then	-- trigger jetpack bursts
			BurstTimer:Reset()
			self.jump = false
		end
		
		if self.BlockingActor then
			if not MovableMan:IsActor(self.BlockingActor) or
				SceneMan:ShortestDistance(self.Pos, self.BlockingActor.Pos, false).Magnitude > (self.Diameter + self.BlockingActor.Diameter)*1.2
			then
				self.BlockingActor = nil
				self.teamBlockState = Actor.NOTBLOCKED
				
				if self.AIMode == Actor.AIMODE_BRAINHUNT and self.FollowingActor then
					self.FollowingActor = nil
					break	-- end this behavior
				end
			elseif self.teamBlockState == Actor.NOTBLOCKED and Waypoint then
				if (Waypoint.Pos.X > self.Pos.X and self.BlockingActor.Pos.X > self.Pos.X) or
					 (Waypoint.Pos.X < self.Pos.X and self.BlockingActor.Pos.X < self.Pos.X)
				then
					self.teamBlockState = Actor.BLOCKED
					if self.AIMode == Actor.AIMODE_BRAINHUNT and self.AIMode == self.BlockingActor.AIMode and
						(not self.BlockingActor.MOMoveTarget or self.BlockingActor.MOMoveTarget.ID ~= self.ID)
					then	-- don't follow an actor that is following us
						self.FollowingActor = self.BlockingActor
						self:ClearAIWaypoints()
						self:AddAIMOWaypoint(self.FollowingActor)
						self:CreateGoToBehavior()
					end
				else
					self.BlockingActor = nil
				end
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	return true
end

-- open fire on the selected target
function HumanBehaviors.ShootTarget(self)
	local aimBonus = 100
	local rateOfFire = 0
	
	if self.EquippedItem and MovableMan:IsActor(self.Target) then
		rateOfFire = ToHDFirearm(self.EquippedItem).RateOfFire
		aimBonus = math.max(aimBonus, ToHeldDevice(self.EquippedItem).SharpLength) / 150	-- make the random error vary with the sharp aim length
	else
		return true
	end
	
	local ShootTimer = Timer()
	local aimTime = RangeRand(600, 1000)
	local TargetAvgVel = self.Target.Vel
	local AimPoint = self.Target.Pos + self.TargetOffset
	local canDuck = false
	
	self.TargetLostTimer:SetSimTimeLimitMS(1500)
	
	-- in air: shoot from the hip
	if self.flying then
		aimTime = aimTime * 0.5
		aimError = RangeRand(-0.4, 0.4)
	else
		aimError = RangeRand(-0.2, 0.2) / aimBonus
		
		if MovableMan:IsActor(self.Target) then
			-- check if we can shoot from the prone position
			canDuck = true
			local x_pos = self.Pos.X
			local ground = self.Pos.Y + self.Height/5
			local Dist = SceneMan:ShortestDistance(self.Pos, AimPoint, false)
			local PronePos
			if Dist.X > 0 then
				PronePos = self.EyePos + Vector(self.Height/4, self.Height/5)
				
				for x = 1, math.floor(self.Height/16+0.5) do
					x_pos = x_pos + 7
					if SceneMan.SceneWrapsX and x_pos > SceneMan.SceneWidth then
						x_pos = SceneMan.SceneWidth - x_pos
					end
					
					if 0 == SceneMan:GetTerrMatter(x_pos, ground) then
						canDuck = false
						break
					end
				end
			else
				PronePos = self.EyePos + Vector(-self.Height/4, self.Height/5)
				
				for x = 1, math.floor(self.Height/16+0.5) do
					x_pos = x_pos - 7
					if SceneMan.SceneWrapsX and x_pos < 0 then
						x_pos = x_pos + SceneMan.SceneWidth
					end
					
					if 0 == SceneMan:GetTerrMatter(x_pos, ground) then
						canDuck = false
						break
					end
				end
			end
			
			-- no LOS from the prone position
			if canDuck then
				Dist = SceneMan:ShortestDistance(PronePos, AimPoint, false)
				if not SceneMan:CastFindMORay(PronePos, Dist, self.Target.ID, Vector(), rte.grassID, false, 4) then
					canDuck = false
				end
			end
			
			if canDuck then
				self.proneState = AHuman.PRONE
				if Dist.X > 0 then
					self.lateralMoveState = Actor.LAT_RIGHT
				else
					self.lateralMoveState = Actor.LAT_LEFT
				end
			end
		end
	end
	
	coroutine.yield()	-- wait until next frame
	
	while true do
		if not MovableMan:IsActor(self.Target) then
			break
		end
		
		if self.FirearmIsReady then
			TargetAvgVel = TargetAvgVel * 0.4 + self.Target.Vel * 0.6	-- filter the target's velocity
			
			if not self.flying then
				self.deviceState = AHuman.AIMING
			end
			
			if SceneMan:ShortestDistance(self.Pos, AimPoint or self.Target.Pos, false).Magnitude < 40 then
				AimPoint = self.Target.Pos + self.TargetOffset * 0.9	-- move the aimpoint closer to the center of the target at close ranges
			else
				AimPoint = self.Target.Pos + self.TargetOffset
			end
			
			-- TODO: replace with weapon specific function
			aim = HumanBehaviors.GetHitAngle(AimPoint, TargetAvgVel/RangeRand(0.2, 4.0), ToHeldDevice(self.EquippedItem).Pos, 90)
			
			if aim then
				self.Ctrl.AnalogAim = Vector(1,0):RadRotate(aim+aimError+RangeRand(-0.018, 0.018))
			else
				break	-- target out of range
			end
			
			if ShootTimer:IsPastSimMS(aimTime) then
				self.fire = true
				
				if ToHDFirearm(self.EquippedItem).FullAuto then
					aimError = aimError * RangeRand(0.89, 0.95)
				else
					ShootTimer:Reset()
					aimTime = rateOfFire + RangeRand(25, 75)
					aimError = aimError * RangeRand(0.7, 0.85)
				end
			else
				self.fire = false
			end
		elseif self:EquipFirearm(true) then
			self.deviceState = AHuman.POINTING
			self.fire = false
			
			ShootTimer:Reset()
			if ToAHuman(self).FirearmIsEmpty then
				self:ReloadFirearm()
				
				-- TODO: check if ducking is appropriate
				
				aimTime = RangeRand(400, 700)
				if self.flying then
					aimTime = aimTime * 0.5
					aimError = RangeRand(-0.28, 0.28)
				else
					aimError = RangeRand(-0.14, 0.14) / aimBonus
				end
			end
		else
			break -- no firearm avaliable
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	return true
end

-- throw a grenade at the selected target
function HumanBehaviors.ThrowTarget(self)
	local ThrowTimer = Timer()
	local aimTime = RangeRand(1000, 1500)
	local scan = 0
	local miss = 0	-- stop scanning after a few missed atempts
	local AimPoint, Dist, MO, ID, rootID, aim, LOS
	
	self.TargetLostTimer:SetSimTimeLimitMS(1500)
	
	while true do
		if not MovableMan:IsActor(self.Target) then
			break
		end
		
		if scan < 1 then
			if self.Target.Door then
				AimPoint = self.Target.Door.Pos
			else
				AimPoint = self.Target.Pos	-- look for the center
				if self.Target.EyePos then
					AimPoint = (AimPoint + self.Target.EyePos) / 2
				end
			end
			
			ID = 255
			if self:IsWithinRange(Vector(AimPoint.X, AimPoint.Y)) then	-- TODO: use grenade properies to decide this
				Dist = SceneMan:ShortestDistance(self.EyePos, AimPoint, false)
				ID = SceneMan:CastMORay(self.EyePos, Dist, self.ID, self.IgnoresWhichTeam, rte.grassID, false, 3)
				if ID < 1 or ID > 254 then	-- not found, look for any head or legs
					AimPoint = self.Target.EyePos	-- the head
					if AimPoint then
						coroutine.yield()	-- wait until next frame
						if not MovableMan:IsActor(self.Target) then	-- must verify that the target exist after a yield
							break
						end
						
						Dist = SceneMan:ShortestDistance(self.EyePos, AimPoint, false)
						ID = SceneMan:CastMORay(self.EyePos, Dist, self.ID, self.IgnoresWhichTeam, rte.grassID, false, 3)
					end
					
					if ID < 1 or ID > 254 then
						local Legs = self.Target.FGLeg or self.Target.BGLeg	-- the legs
						if Legs then
							coroutine.yield()	-- wait until next frame
							if not MovableMan:IsActor(self.Target) then	-- must verify that the target exist after a yield
								break
							end
							
							AimPoint = Legs.Pos
							Dist = SceneMan:ShortestDistance(self.EyePos, AimPoint, false)
							ID = SceneMan:CastMORay(self.EyePos, Dist, self.ID, self.IgnoresWhichTeam, rte.grassID, false, 3)
						end
					end
				end
			else
				break	-- out of range
			end
			
			if ID > 0 and ID < 255 then	-- MO found
				scan = 6	-- skip the LOS check the next n frames
				miss = 0
				LOS = true	-- we have line of sight to the target
				
				-- check what target we will hit
				rootID = MovableMan:GetRootMOID(ID)
				if rootID ~= self.Target.ID then
					MO = MovableMan:GetMOFromID(rootID)
					if MovableMan:IsActor(MO) then
						if MO.Team ~= self.Team then
							if MO.ClassName == "AHuman" then
								self.Target = ToAHuman(MO)
							elseif MO.ClassName == "ACrab" then
								self.Target = ToACrab(MO)
							elseif MO.ClassName == "ACRocket" then
								self.Target = ToACRocket(MO)
							elseif MO.ClassName == "ACDropShip" then
								self.Target = ToACDropShip(MO)
							elseif MO.ClassName == "ADoor" then
								self.Target = ToADoor(MO)
							elseif MO.ClassName == "Actor" then
								self.Target = ToActor(MO)
							else
								break
							end
						else
							break	-- don't shoot friendlies
						end
					end
				end
			else
				miss = miss + 1
				if miss > 4 then	-- stop looking if we cannot find anything after n atempts
					break
				else
					scan = 3	-- check LOS a little bit more often if no MO was found
				end
			end
			
			if LOS then	-- don't sharp aim until LOS has been confirmed
				if self.ThrowableIsReady then
					aim = HumanBehaviors.GetHitAngle(AimPoint, self.Target.Vel/RangeRand(0.5, 1.75), ToHeldDevice(self.EquippedItem).MuzzlePos, 18)
					if aim then
						self.Ctrl.AnalogAim = Vector(1,0):RadRotate(aim+RangeRand(-0.1, 0.1))
					else
						break	-- target out of range
					end
					
					if not ThrowTimer:IsPastSimMS(aimTime) then
						self.fire = true
					else
						ThrowTimer:Reset()
						self.fire = false
						aimTime = RangeRand(1000, 1500)
					end
				else
					break	-- no grenades left
				end
			end
		else
			scan = scan - 1
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	return true
end

-- attack the target in hand-to-hand
function HumanBehaviors.AttackTarget(self)
	if not self.Target or not MovableMan:IsActor(self.Target) then
		return true
	end
	
	self.TargetLostTimer:SetSimTimeLimitMS(5000)
	
	local OldWaypoint
	if not self.MOMoveTarget or 
		not MovableMan:ValidMO(self.MOMoveTarget) or
		self.MOMoveTarget.RootID ~= self.Target.RootID
	then
		OldWaypoint = self:GetLastAIWaypoint()
		if OldWaypoint.Magnitude == 0 then
			OldWaypoint = nil
		end
		
		self:ClearAIWaypoints()
		self:AddAIMOWaypoint(self.Target)
		self:CreateGoToBehavior()
	end
	
	while true do
		coroutine.yield()	-- wait until next frame
		
		if not self.Target or not MovableMan:IsActor(self.Target) then
			self:ClearAIWaypoints()
			if OldWaypoint then	-- move towards the old wpt
				self:AddAISceneWaypoint(OldWaypoint)
				self:CreateGoToBehavior()
			end

			break
		end
		
		if self.EquippedItem and ToHeldDevice(self.EquippedItem):IsTool() then	-- attack with digger
			local Dist = SceneMan:ShortestDistance(ToHeldDevice(self.EquippedItem).MuzzlePos, self.Target.Pos, false)
			if Dist.Magnitude < 40 then
				self.Ctrl.AnalogAim = SceneMan:ShortestDistance(self.EyePos, self.Target.Pos, false).Normalized
				self.fire = true
			else
				self.fire = false
			end
		end
	end
	
	return true
end


-- move around another actor
function HumanBehaviors.MoveAroundActor(self)
	if not MovableMan:IsActor(self.BlockingActor) then
		self.teamBlockState = Actor.NOTBLOCKED
		self.BlockingActor = nil
		return true
	end
	
	local BurstTimer = Timer()
	local refuel = false
	local Dist
	
	BurstTimer:SetSimTimeLimitMS(200)	-- a burst last until the BurstTimer expire
	self.jump = true
	
	-- look above the blocking actor
	Dist = SceneMan:ShortestDistance(self.Pos, self.BlockingActor.Pos, false)
	if Dist.X > 0 then
		self.Ctrl.AnalogAim = Vector(1,0):RadRotate(1.20)
	else
		self.Ctrl.AnalogAim = Vector(1,0):RadRotate(1.94)
	end
	
	while true do		
		if BurstTimer:IsPastSimTimeLimit() then	-- trigger jetpack bursts
			BurstTimer:Reset()
			self.jump = false
			
			Dist = SceneMan:ShortestDistance(self.Pos, self.BlockingActor.Pos, false)
			if Dist.Y + self.Vel.Y * 3 > (self.Diameter + self.BlockingActor.Diameter)*0.67 then
				self:SetAimAngle(-0.5)
				
				if math.abs(Dist.X) > math.max(self.Diameter, self.BlockingActor.Diameter)/2 then
					return true
				end
			end
		else
			self.jump = true
			if self.Vel.Y < -9 then
				self.jump = false
			end
		end
		
		if refuel then
			self.jump = false
			if self.JetTimeLeft > self.JetTimeTotal * 0.9 then
				refuel = false
			end
		elseif self.JetTimeLeft < self.JetTimeTotal * 0.1 then
			refuel = true
		end
		
		coroutine.yield()	-- wait until next frame
		if not MovableMan:IsActor(self.BlockingActor) then
			self.teamBlockState = Actor.NOTBLOCKED
			self.BlockingActor = nil
			return true
		end
	end
	
	return true
end

-- stop the user fom inadvertently modifying the storage table
local Proxy = {}
local Mt = {
	__index = HumanBehaviors,
	__newindex = function(Table, k, v)
		error("The HumanBehaviors table is read-only.", 2)
	end
}
setmetatable(Proxy, Mt)
HumanBehaviors = Proxy
