package.loaded.Constants = nil; require("Constants")
require("Actors/AI/TurretBehaviors")

function Create(self)
	---------------- AI variables start ----------------
	self.Ctrl = self:GetController()
	
	self.lateralMoveState = Actor.LAT_STILL
	self.deviceState = ACrab.STILL
	self.lastAIMode = Actor.AIMODE_NONE
	self.SentryFacing = self.HFlipped
	self.fire = false
	
	self.ReloadTimer = Timer()
	self.TargetLostTimer = Timer()
	
	self.PlayerInterferedTimer = Timer()
	self.PlayerInterferedTimer:SetSimTimeLimitMS(500)
	
	self.AlarmTimer = Timer()
	self.AlarmTimer:SetSimTimeLimitMS(500)
	
	-- check if this team is controlled by a human
	local Activ = ActivityMan:GetActivity()
	for player = Activity.PLAYER_1, Activity.MAXPLAYERCOUNT - 1 do
		if Activ:PlayerActive(player) and Activ:PlayerHuman(player) then
			if self.Team == Activ:GetTeamOfPlayer(player) then
				self.isPlayerOwned = true
				break
			end
		end
	end
	
	function self:LookForTargets()
		local viewAng
		
		if self.FirearmIsReady then
			viewAng = 0.1
		else
			viewAng = 0.4
		end
		
		local viewLen = SceneMan:ShortestDistance(self.EyePos, self.ViewPoint, false).Magnitude + FrameMan.PlayerScreenWidth*0.5
		local Trace = Vector(viewLen, 0):RadRotate(viewAng*NormalRand()+self:GetAimAngle(true))
		local ID = SceneMan:CastMORay(self.EyePos, Trace, self.ID, self.IgnoresWhichTeam, rte.grassID, false, 5)
		
		if ID < 255 then
			local Point = SceneMan:GetLastRayHitPos()
			if not SceneMan:IsUnseen(Point.X, Point.Y, self.Team) then
				return MovableMan:GetMOFromID(ID)
			end
		end
		
		return nil
	end
	
	function self:CreatePatrolBehavior()
		self.NextBehavior = coroutine.create(TurretBehaviors.Patrol)
		self.NextCleanup = nil
		self.NextBehaviorName = "Patrol"
	end
	
	function self:CreateSentryBehavior()
		if not self.FirearmIsReady and not self.ThrowableIsReady then
			return
		end
		
		self.NextBehavior = coroutine.create(TurretBehaviors.Sentry)
		self.NextCleanup = nil
		self.NextBehaviorName = "Sentry"
	end
	
	function self:CreateAttackBehavior()
		if self.FirearmIsReady then
			self.NextBehavior = coroutine.create(TurretBehaviors.ShootTarget)
			self.NextBehaviorName = "ShootTarget"
		else
			if self.FirearmIsEmpty then
				self:ReloadFirearm()
			end
			
			return
		end
		
		self.NextCleanup = function(self)
			self.fire = false
			self.Target = nil
			self.deviceState = ACrab.STILL
		end
	end
	
	---------------- AI variables end ----------------
end

function UpdateAI(self)
	if self.PlayerInterferedTimer:IsPastSimTimeLimit() then
		self.Behavior = nil	-- remove the current behavior
		if self.BehaviorCleanup then
			self.BehaviorCleanup(self)	-- clean up after the current behavior
			self.BehaviorCleanup = nil
		end
		
		self.Target = nil
		self.fire = false
		self.lastAIMode = Actor.AIMODE_NONE
	end
	
	self.PlayerInterferedTimer:Reset()
	
	-- switch to the next behavior, if avaliable
	if self.NextBehavior then
		if self.BehaviorCleanup then
			self.BehaviorCleanup(self)
		end
		
		self.Behavior = self.NextBehavior
		self.BehaviorCleanup = self.NextCleanup
		self.BehaviorName = self.NextBehaviorName
		self.NextBehavior = nil
	end
	
	-- check if the AI mode has changed or if we need a new behavior
	if self.AIMode ~= self.lastAIMode or not self.Behavior then
		self.Behavior = nil
		if self.BehaviorCleanup then
			self.BehaviorCleanup(self)	-- stop the current behavior
			self.BehaviorCleanup = nil
		end
		
		-- select a new behavior based on AI mode
		if self.AIMode == Actor.AIMODE_PATROL then
			self:CreatePatrolBehavior()
		else
			if self.AIMode ~= self.lastAIMode and self.AIMode == Actor.AIMODE_SENTRY then
				self.SentryFacing = self.HFlipped	-- store the direction in which we should be looking
			end
			
			self:CreateSentryBehavior()
		end
		
		self.lastAIMode = self.AIMode
	end
	
	-- look for targets
	local FoundMO = self:LookForTargets()
	if FoundMO then
		if self.Target and MovableMan:IsActor(self.Target) and FoundMO.RootID == self.Target.RootID then	-- found the same target
			local Point = SceneMan:GetLastRayHitPos()
			self.TargetOffset = (self.TargetOffset + SceneMan:ShortestDistance(self.Target.Pos, Point, false))/2
			self.ReloadTimer:Reset()
			if not SceneMan:IsUnseen(Point.X, Point.Y, self.Team) then
				self.TargetLostTimer:Reset()
			end
		else
			FoundMO = MovableMan:GetMOFromID(FoundMO.RootID)
			if MovableMan:IsActor(FoundMO) then
				FoundMO = ToActor(FoundMO)
				
				if FoundMO.Team == self.Team then
					self.Target = nil	-- stop shooting
				else
					if FoundMO.ClassName == "AHuman" then
						FoundMO = ToAHuman(FoundMO)
					elseif FoundMO.ClassName == "ACrab" then
						FoundMO = ToACrab(FoundMO)
					elseif FoundMO.ClassName == "ACRocket" then
						FoundMO = ToACRocket(FoundMO)
					elseif FoundMO.ClassName == "ACDropShip" then
						FoundMO = ToACDropShip(FoundMO)
					elseif FoundMO.ClassName == "ADoor" then
						FoundMO = ToADoor(FoundMO)
					elseif FoundMO.ClassName ~= "Actor" then
						FoundMO = nil
					end
					
					if FoundMO then
						self.ReloadTimer:Reset()
						
						if not self.Target or not MovableMan:IsActor(self.Target) then
							self.OldTargetPos = nil
							self.Target = FoundMO
							self.TargetOffset = SceneMan:ShortestDistance(self.Target.Pos, SceneMan:GetLastRayHitPos(), false)	-- this it the distance vector from the target center to the point we hit with our ray
							self.TargetLostTimer:SetSimTimeLimitMS(1000)
							self:CreateAttackBehavior()
						--else
							-- TODO: check if this target have higher priority
						end
						
						local Point = self.Target.Pos + self.TargetOffset
						if not SceneMan:IsUnseen(Point.X, Point.Y, self.Team) then
							self.TargetLostTimer:Reset()
						end
					end
				end
			end
		end
	else
		if self.Target then
			if MovableMan:IsActor(self.Target) then
				if self.TargetLostTimer:IsPastSimTimeLimit() then
					self.OldTargetPos = self.Target.Pos
					self.Target = nil	-- the target has been out of sight for too long, ignore it
				end
			else
				self.Target = nil
			end
		end
		
		if self.ReloadTimer:IsPastSimMS(8000) then	-- check if we need to reload
			self.ReloadTimer:Reset()
			if ToACrab(self).FirearmNeedsReload then
				self:ReloadFirearm()
			end
		end
	end
	
	-- run the selected behavior and delete it if it returns true
	if self.Behavior then
		local msg, done = coroutine.resume(self.Behavior, self)
		if not msg then
			ConsoleMan:PrintString(self.PresetName .. " behavior " .. self.BehaviorName .. " error:\n" .. done)	-- print the error message
			done = true
		end
		
		if done then
			self.Behavior = nil
			if self.BehaviorCleanup then
				self.BehaviorCleanup(self)
				self.BehaviorCleanup = nil
			end
		end
	end
	
	-- react to relevant AlarmEvents
	if self.AlarmTimer:IsPastSimTimeLimit() then
		self.LastAlarmPos = self.AlarmPos
		self.AlarmPos = nil
		
		for Event in MovableMan.AlarmEvents do
			if Event.Team ~= self.Team then	-- caused by some other team's activites - alarming!
				local AlarmVec = SceneMan:ShortestDistance(self.EyePos, Event.ScenePos, false)	-- see how far away the alarm situation is
				
				-- only react if the alarm is within range
				if AlarmVec.Largest * self.Perceptiveness <= Event.Range then
					-- if this is the same alarm location as last, then don't repeat the signal
					if not self.LastAlarmPos or SceneMan:ShortestDistance(self.LastAlarmPos, Event.ScenePos, false).Largest > 10 then
						-- now check if we have line of sight to the alarm point
						if not SceneMan:CastStrengthRay(self.EyePos, AlarmVec, 6, Vector(), 7, rte.grassID, true) then
							self.AlarmPos = Event.ScenePos
							self.AlarmTimer:Reset()
							break
						end
					end
				end
			end
		end
	elseif self.AlarmPos then	-- if alarmed, look at the alarming point until AlarmTimer has expired
		-- point in the direciton we heard the alarm come from if we're not already engaging a target
		if not (self.Target and self.deviceState == ACrab.FIRING) then
			-- aim in the direction of alarm
			self.deviceState = ACrab.AIMING
			self.Ctrl.AnalogAim = SceneMan:ShortestDistance(self.EyePos, self.AlarmPos, false).Normalized
		end
	end
	
	-- controller states
	if self.fire then
		self.Ctrl:SetState(Controller.WEAPON_FIRE, true)
	end
	
	if self.deviceState == ACrab.AIMING then
		self.Ctrl:SetState(Controller.AIM_SHARP, true)
	end
end
