package.loaded.Constants = nil; require("Constants")

TurretBehaviors = {}

function TurretBehaviors.Sentry(self)	-- in sentry behavior the agent only looks for new enemies, it sometimes sharp aims to increse spotting range
	local sweepUp = true
	local sweepDone = false
	local maxAng = 0.49	-- TODO: get AimRange from the ini
	local minAng = -0.49
	local aim
	
	if self.OldTargetPos then	-- try to reaquire an old target
		local Dist = SceneMan:ShortestDistance(self.EyePos, self.OldTargetPos, false)
		self.OldTargetPos = nil
		if (Dist.X < 0 and self.HFlipped) or (Dist.X > 0 and not self.HFlipped) then	-- we are facing the target	
			self.deviceState = ACrab.AIMING
			self.Ctrl.AnalogAim = Dist.Normalized
			
			for i = 1, 30 do
				coroutine.yield()	-- aim here for ~0.5s
			end
		end
	elseif not self.isPlayerOwned then -- face the most likely enemy approach direction
		for i = 1, math.random(5) do	-- wait for a while
			coroutine.yield()
		end
		
		self:AddAISceneWaypoint(Vector(self.Pos.X, 0))
		self:UpdateMovePath()
		coroutine.yield()	-- wait until next frame
		
		-- face the direction of the first waypoint
		for WptPos in self.MovePath do
			local Dist = SceneMan:ShortestDistance(self.Pos, WptPos, false)
			if Dist.X > 5 then
				self.SentryFacing = false
				self.Ctrl.AnalogAim = Dist.Normalized
			elseif Dist.X < -5 then
				self.SentryFacing = true
				self.Ctrl.AnalogAim = Dist.Normalized
			end
			
			break
		end
	end
	
	while true do	-- start by looking forward
		aim = self:GetAimAngle(false)
		
		if sweepUp then
			if aim < maxAng/3 then
				self.Ctrl:SetState(Controller.AIM_UP, false)
				coroutine.yield()	-- wait until next frame
				self.Ctrl:SetState(Controller.AIM_UP, true)
			else
				sweepUp = false
			end
		else
			if aim > minAng/3 then
				self.Ctrl:SetState(Controller.AIM_DOWN, false)
				coroutine.yield()	-- wait until next frame
				self.Ctrl:SetState(Controller.AIM_DOWN, true)
			else
				sweepUp = true
				if sweepDone then
					break
				else
					sweepDone = true
				end
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	if self.HFlipped ~= self.SentryFacing then
		self.HFlipped = self.SentryFacing	-- turn to the direction we have been order to guard
		return true	-- restart this behavior
	end
	
	while true do	-- look down
		aim = self:GetAimAngle(false)
		if aim > minAng then
			self.Ctrl:SetState(Controller.AIM_DOWN, true)
		else
			break
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	local Hit = Vector()
	local NoObstacle = {}
	self.deviceState = ACrab.AIMING
	
	while true do	-- scan the area for obstacles
		aim = self:GetAimAngle(false)
		if aim < maxAng then
			self.Ctrl:SetState(Controller.AIM_UP, true)
		else
			break
		end
		
		-- save the angle to a table if there is no obstacle
		if not SceneMan:CastStrengthRay(self.EyePos, Vector(60, 0):RadRotate(self:GetAimAngle(true)), 5, Hit, 2, 0, true) then
			table.insert(NoObstacle, aim)	-- TODO: don't use a table for this
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	local SharpTimer = Timer()
	local aimTime = 2000
	local angDiff = 1
	self.deviceState = ACrab.POINTING
	
	if #NoObstacle > 1 then	-- only aim where we know there are no obstacles, e.g. out of a gun port
		minAng = NoObstacle[1] * 0.95
		maxAng = NoObstacle[#NoObstacle] * 0.95
		angDiff = 1 / math.max(math.abs(maxAng - minAng), 0.1)	-- sharp aim longer from a small aiming window
	end
	
	while true do
		aim = self:GetAimAngle(false)
		
		if sweepUp then
			if aim < maxAng then
				if aim < maxAng/5 and aim > minAng/5 and PosRand() > 0.3 then
					self.Ctrl:SetState(Controller.AIM_UP, false)
				else
					self.Ctrl:SetState(Controller.AIM_UP, true)
				end
			else
				sweepUp = false
			end
		else
			if aim > minAng then
				if aim < maxAng/5 and aim > minAng/5 and PosRand() > 0.3 then
					self.Ctrl:SetState(Controller.AIM_DOWN, false)
				else
					self.Ctrl:SetState(Controller.AIM_DOWN, true)
				end
			else
				sweepUp = true
			end
		end
		
		if SharpTimer:IsPastSimMS(aimTime) then
			SharpTimer:Reset()
			
			if self.deviceState == ACrab.AIMING then
				aimTime = RangeRand(1000, 3000)
				self.deviceState = ACrab.POINTING
			else
				aimTime = RangeRand(6000, 12000) * angDiff
				self.deviceState = ACrab.AIMING
			end
			
			if self.HFlipped ~= self.SentryFacing then
				self.HFlipped = self.SentryFacing	-- turn to the direction we have been order to guard
				break	-- restart this behavior
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	return true
end



function TurretBehaviors.Patrol(self)
	local TurnTimer = Timer()
	local turn = 3000
	
	while true do
		if TurnTimer:IsPastSimMS(turn) then
			TurnTimer:Reset()
			turn = RangeRand(2000, 6000)
			self:SetAimAngle(RangeRand(-0.7, 0.7))
			
			self.HFlipped = not self.HFlipped
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	return true
end


-- open fire on the selected target
function TurretBehaviors.ShootTarget(self)
	if not MovableMan:IsActor(self.Target) then
		return true
	end
	
	local TargetAvgVel = self.Target.Vel
	local ShootTimer = Timer()
	local aimTime = RangeRand(600, 1000)
	local aimError = RangeRand(-0.2, 0.2)
	local aim, AimPoint
	
	while true do		
		if self.FirearmIsReady then
			TargetAvgVel = TargetAvgVel * 0.4 + self.Target.Vel * 0.6	-- filter the target's velocity
			
			if not self.flying then
				self.deviceState = ACrab.AIMING
			end
			
			if SceneMan:ShortestDistance(self.Pos, AimPoint or self.Target.Pos, false).Magnitude < 40 then
				AimPoint = self.Target.Pos + self.TargetOffset * 0.9	-- move the aimpoint closer to the center of the target at close ranges
			else
				AimPoint = self.Target.Pos + self.TargetOffset
			end
			
			aim = SceneMan:ShortestDistance(self.EyePos, AimPoint, false).AbsRadAngle
			self.Ctrl.AnalogAim = Vector(1,0):RadRotate(aim+aimError+RangeRand(-0.018, 0.018))
			
			if ShootTimer:IsPastSimMS(aimTime) then
				self.fire = true
				aimError = aimError * RangeRand(0.95, 0.99)
			else
				self.fire = false
			end
		else
			self.deviceState = ACrab.POINTING
			self.fire = false
			
			ShootTimer:Reset()
			if self.FirearmIsEmpty then
				self:ReloadFirearm()
				aimTime = RangeRand(400, 700)
				aimError = RangeRand(-0.14, 0.14)
			else
				break -- no firearm avaliable
			end
		end
		
		coroutine.yield()	-- wait until next frame
		
		if not MovableMan:IsActor(self.Target) then
			break
		end
	end
	
	return true
end

-- stop the user fom inadvertently modifying the storage table
local Proxy = {}
local Mt = {
	__index = TurretBehaviors,
	__newindex = function(Table, k, v)
		error("The TurretBehaviors table is read-only.", 2)
	end
}
setmetatable(Proxy, Mt)
TurretBehaviors = Proxy
