require("Actors/AI/CrabAI")

function Update(self)
--Just here to see notifications of this running.
--print("running the movement script!")

--Create the necessary variables.
if self.timer == nil then
	self.timer = Timer();
end

--Start our loop, every 10-20ms.
local randomTimer = math.random(10,20);
if self.timer:IsPastSimMS(randomTimer) then
	local p = Vector();

	--Core of the Movement script starts here.
	--Important bit here- adjust the altitude value until it's stable and not too floaty.  
	--This specific altitude should be slightly lower than the AHuman's hip center to the ground.
	local altCheck = 9
	local heightCheck = 35
	local sideCheck = 15
	local upPush = 0.5
	local downPush = 1.5
	local downCheck = false
	local upCheck = false

	local centerAlt = SceneMan:CastObstacleRay(self.Pos, Vector(0,altCheck), p, p, self.ID, 0, 0, 0);
	local frontAlt = SceneMan:CastObstacleRay(Vector(self.Pos.X-4,self.Pos.Y), Vector(0,altCheck), p, p, self.ID, 0, 0, 0);
	local backAlt = SceneMan:CastObstacleRay(Vector(self.Pos.X+4,self.Pos.Y), Vector(0,altCheck), p, p, self.ID, 0, 0, 0);
	
	--Safety code, because we get -1 if no obstacles
	if centerAlt < 0 then
		centerAlt = 1000
	end
	if frontAlt < 0 then
		frontAlt = 1000
	end
	if backAlt < 0 then
		backAlt = 1000
	end

	--Now that we know the heights to obstacles, it's time to perform some logic.
	--May need to alter the velPushfor best performance depending on mass.
	--Please note that this allows for ducking and prone and even movement while body is lowered; let go of the control...
	--...and the character "stands" back up again.
	if (centerAlt < altCheck or frontAlt < altCheck or backAlt < altCheck) and not self:GetController():IsState(Controller.MOVE_DOWN) then
		downCheck = true
	elseif self:GetController():IsState(Controller.MOVE_DOWN) then
		self.Vel.Y = self.Vel.Y + 0.1
	end

	--Do a raytest vs. hitting top of the Actor on things.  If buried, push down, hard.
	local ray = SceneMan:CastObstacleRay(self.Pos, Vector(4,-heightCheck), p, p, self.ID, 0, 0, 0);
	if ray < heightCheck and ray ~= -1 then
		upCheck = true
	end
	ray = SceneMan:CastObstacleRay(self.Pos, Vector(-4,-heightCheck), p, p, self.ID, 0, 0, 0);
	if ray < heightCheck and ray ~= -1 then
		upCheck = true
	end
	
	--If we're supposed to push DOWN or UP, do one or the other, but if both conditions... 
	--...are true or false, do NOTHING.  Prevents odd behaviors underground
	if downCheck == true and upCheck == false then
		self.Vel.Y = self.Vel.Y - upPush
	end	
	if downCheck == false and upCheck == true then
		self.Vel.Y = self.Vel.Y + downPush
	end

	--Do raytests to the right and left to make sure we're not running through walls.
	--This isn't exactly super-reliable, but it's better'n nothin'.
	local rightLeft = 0
	ray = SceneMan:CastObstacleRay(self.Pos, Vector(sideCheck,0), p, p, self.ID, 0, 0, 0);
	if ray < sideCheck and ray ~= -1 then
		rightLeft = rightLeft -1
	end
	ray = SceneMan:CastObstacleRay(self.Pos, Vector(-sideCheck,0), p, p, self.ID, 0, 0, 0);
	if ray < sideCheck and ray ~= -1 then
		rightLeft = rightLeft + 1
	end
	--If we want to push right OR left, but NOT push right AND left, then push right or left.
	--Sorry for the logic joke.
	if rightLeft == -1 then
		self.Vel.X = self.Vel.X - 1.0
	end
	if rightLeft == 1 then
		self.Vel.X = self.Vel.X + 1.0
	end	

	--Brute-force change in velocity over time. 
	--If accel is set really high, bad things will happen ;-)
	local speed_max = 6.0
	local accel = 0.5
	if self:GetController():IsState(Controller.MOVE_LEFT) and rightLeft == 0 then
		self.Vel.X = math.max(self.Vel.X - accel,-speed_max)
	end
	if self:GetController():IsState(Controller.MOVE_RIGHT) and rightLeft == 0 then
		self.Vel.X = math.min(self.Vel.X + accel,speed_max)
	end	
	
	--Rotational control; provides some rotational force over time to raise/lower the body.
	--Note that the back control is stronger than the front, because of some lame body-lean...
	--...that is apparently hard-coded into ACrab skeletal sim :-/
	if self.HFlipped then
		--print("flipped")
		if math.abs(frontAlt - backAlt) < 1 then
			if frontAlt > backAlt then
				--print("front is higher than back")
				self.AngularVel = math.min(self.AngularVel + 0.05,0.25)
			end
			
			if frontAlt < backAlt then
				--print("back is higher than front")
				self.AngularVel = math.max(self.AngularVel - 0.1,-0.375)
			end
		else
			if frontAlt > backAlt then
				--print("front is higher than back")
				self.AngularVel = math.min(self.AngularVel + 0.1,0.5)
			end
			
			if frontAlt < backAlt then
				--print("back is higher than front")
				self.AngularVel = math.max(self.AngularVel - 0.2,-0.75)
			end
		end
	else
		--print("not flipped")
		if math.abs(frontAlt - backAlt) < 1 then		
			if frontAlt > backAlt then
				--print("front is higher than back")
				self.AngularVel = math.min(self.AngularVel + 0.1,0.375)
			end
			
			if frontAlt < backAlt then
				--print("back is higher than front")
				self.AngularVel = math.max(self.AngularVel - 0.05,-0.25)
			end		
		else
			if frontAlt > backAlt then
				--print("front is higher than back")
				self.AngularVel = math.min(self.AngularVel + 0.2,0.75)
				
			end
			
			if frontAlt < backAlt then
				--print("back is higher than front")
				self.AngularVel = math.max(self.AngularVel - 0.1,-0.5)
			end			
		end	
	end
	
	--Brute-force adjustments of RotAngles, to prevent funky behaviors.
	if self.RotAngle > 0.4 then
		self.RotAngle = 0.4
	end	
	if self.RotAngle < - 0.4 then
		self.RotAngle = - 0.4
	end		
	--It's flatter, use less RotAngle.
	if math.abs(frontAlt - backAlt) < 10 then
		if self.RotAngle > 0.2 then
			self.RotAngle = 0.2
		end	
		if self.RotAngle < - 0.2 then
			self.RotAngle = - 0.2
		end		
	end		
	--It's really flat, do not allow rotational forces to screw it up.
	if math.abs(frontAlt - backAlt) < 5 then
		if self.RotAngle > 0.1 then
			self.RotAngle = 0.1
		end	
		if self.RotAngle < - 0.1 then
			self.RotAngle = - 0.1
		end		
	end

	--Brute-force removal of "falling through the earth"; troops hit "bedrock".
	if self.Pos.Y > SceneMan.SceneHeight - 40 then 
		self.Pos.Y = SceneMan.SceneHeight - 42
		self.Vel.Y = -1.0
	end

	--Reset our 10ms timer
	self.timer:Reset();
--End our loop
end
--End the Update() cycle
end