require("Actors/AI/HumanAI")

function Update(self)
--Just here to see notifications of this running.
--print("running the movement script!")

--Create the necessary variables.
if self.timer == nil then
	self.timer = Timer();
    --Keep track of how long it should be before healing people.
    self.healTimer = Timer();
    --Interval between healings, in milliseconds.
    self.healInterval = 100;
    --Heal counter.
    self.healnum = 0;
end

--Start our loop, every 10-20ms.
local randomTimer = math.random(10,20);
if self.timer:IsPastSimMS(randomTimer) then
	local p = Vector();

	--Core of the Movement script starts here.
	--Important bit here- adjust the altitude value until it's stable and not too floaty.  
	--This specific altitude should be slightly lower than the AHuman's hip center to the ground.
	local altCheck = 19
	local heightCheck = 30
	local sideCheck = 10
	local upPush = 0.35
	local downPush = 0.7
	local downCheck = false
	local upCheck = false

	local centerAlt = SceneMan:CastObstacleRay(self.Pos, Vector(0,altCheck), p, p, self.ID, 0, 0, 0);
	local frontAlt = SceneMan:CastObstacleRay(Vector(self.Pos.X-4,self.Pos.Y), Vector(0,altCheck), p, p, self.ID, 0, 0, 0);
	local backAlt = SceneMan:CastObstacleRay(Vector(self.Pos.X+4,self.Pos.Y), Vector(0,altCheck), p, p, self.ID, 0, 0, 0);
	
	--Safety code, because we get -1 if no obstacles
	if centerAlt < 0 then
		centerAlt = 1000
	end
	if frontAlt < 0 then
		frontAlt = 1000
	end
	if backAlt < 0 then
		backAlt = 1000
	end

	--Now that we know the heights to obstacles, it's time to perform some logic.
	--May need to alter the velPushfor best performance depending on mass.
	--Please note that this allows for ducking and prone and even movement while body is lowered; let go of the control...
	--...and the character "stands" back up again.
	if (centerAlt < altCheck or frontAlt < altCheck or backAlt < altCheck) and not self:GetController():IsState(Controller.MOVE_DOWN) then
		downCheck = true
	elseif self:GetController():IsState(Controller.MOVE_DOWN) then
		self.Vel.Y = self.Vel.Y + 0.1
	end

	--Do a raytest vs. hitting top of the Actor on things.  If buried, push down, hard.
	local ray = SceneMan:CastObstacleRay(self.Pos, Vector(4,-heightCheck), p, p, self.ID, 0, 0, 0);
	if ray < heightCheck and ray ~= -1 then
		upCheck = true
	end
	ray = SceneMan:CastObstacleRay(self.Pos, Vector(-4,-heightCheck), p, p, self.ID, 0, 0, 0);
	if ray < heightCheck and ray ~= -1 then
		upCheck = true
	end
	
	--If we're supposed to push DOWN or UP, do one or the other, but if both conditions... 
	--...are true or false, do NOTHING.  Prevents odd behaviors underground
	if downCheck == true and upCheck == false then
		self.Vel.Y = self.Vel.Y - upPush
	end	
	if downCheck == false and upCheck == true then
		self.Vel.Y = self.Vel.Y + downPush
	end

	--Do raytests to the right and left to make sure we're not running through walls.
	--This isn't exactly super-reliable, but it's better'n nothin'.
	local rightLeft = 0
	ray = SceneMan:CastObstacleRay(self.Pos, Vector(sideCheck,0), p, p, self.ID, 0, 0, 0);
	if ray < sideCheck and ray ~= -1 then
		rightLeft = rightLeft -1
	end
	ray = SceneMan:CastObstacleRay(self.Pos, Vector(-sideCheck,0), p, p, self.ID, 0, 0, 0);
	if ray < sideCheck and ray ~= -1 then
		rightLeft = rightLeft + 1
	end
	--If we want to push right OR left, but NOT push right AND left, then push right or left.
	--Sorry for the logic joke.
	if rightLeft == -1 then
		self.Vel.X = self.Vel.X - 1.0
	end
	if rightLeft == 1 then
		self.Vel.X = self.Vel.X + 1.0
	end	

	--Brute-force change in velocity over time. 
	--If accel is set really high, bad things will happen ;-)
	local speed_max = 8.0
	local accel = 1.0
	if self:GetController():IsState(Controller.MOVE_LEFT) and rightLeft == 0 then
		self.Vel.X = math.max(self.Vel.X - accel,-speed_max)
	end
	if self:GetController():IsState(Controller.MOVE_RIGHT) and rightLeft == 0 then
		self.Vel.X = math.min(self.Vel.X + accel,speed_max)
	end	
	
	--Brute-force adjustments of RotAngles, to prevent funky behaviors.
	if self.RotAngle > 0.2 then
		self.RotAngle = 0.2
	end	
	if self.RotAngle < - 0.2 then
		self.RotAngle = - 0.2
	end

	--Brute-force removal of "falling through the earth"; troops hit "bedrock".
	if self.Pos.Y > SceneMan.SceneHeight - 40 then 
		self.Pos.Y = SceneMan.SceneHeight - 42
		self.Vel.Y = -1.0
	end		

	--Reset our 10ms timer
	self.timer:Reset();
--End our loop
end

    local avgx = nil;
    local avgy = nil;
    local dist = nil;
    local part = nil;
    --Heal every interval.
    if self.healTimer:IsPastSimMS(self.healInterval) then
	--Cycle through all actors.
	for actor in MovableMan.Actors do
	    --If the actor is on the right team, has less than 100 health and is not the healer, continue with code.
	    if actor.Team == self.Team and actor.Health < 100 and actor.ID ~= self.ID then
		--Trigonometry to find how far the actor is.
		avgx = actor.Pos.X - self.Pos.X;
		avgy = actor.Pos.Y - self.Pos.Y;
		dist = math.sqrt(avgx ^ 2 + avgy ^ 2);
		if dist < 200 then
		    --If the actor is fairly close, heal them!
		    actor.Health = actor.Health + 1;
		    --Every 8 health healed, put a little icon above the actor's head to show that it is indeed healing.
		    if self.healnum == 8 then
		    	--Create the particle.
		    	part = CreateMOSParticle("Particle Heal Effect");
		    	--Set the particle's position to just over the actor's head.
		    	part.Pos = actor.AboveHUDPos + Vector(0,4);
			--Add the particle to the world.
		    	MovableMan:AddParticle(part);
		    end
		end
	    end
	end
	--Reset the healing timer.
	self.healTimer:Reset();
	--Increment the heal counter.
	self.healnum = self.healnum + 1;
	--Reset it if it's gone too far.
	if self.healnum > 8 then
	    self.healnum = 0;
	end
    end

--End the Update() cycle
end